@@ -540,7 +540,7 @@ format_process_exename (void *data, char *&destbuf)
     stpcpy (buf, "<defunct>");
   else
     {
-      mount_table->conv_to_posix_path (p->progname, buf, 1);
+      mount_table->conv_to_posix_path (p->progname, buf, CCP_RELATIVE);
       len = strlen (buf);
       if (len > 4)
 	{
@@ -648,7 +648,7 @@ struct heap_info
 	  stpcpy (p, "]");
 	  return dest;
 	}
-    return 0;
+    return NULL;
   }
 
   ~heap_info ()
@@ -731,7 +731,8 @@ struct thread_info
 	  {
 	    *r = (region) { regions, (ULONG) (ULONG_PTR) thread[i].ClientId.UniqueThread,
 			    (char *) tbi.TebBaseAddress,
-			    (char *) tbi.TebBaseAddress + wincap.page_size (),
+			    (char *) tbi.TebBaseAddress
+				     + 2 * wincap.page_size (),
 			    true };
 	    regions = r;
 	  }
@@ -769,7 +770,18 @@ struct thread_info
 	  stpcpy (p, "]");
 	  return dest;
 	}
-    return 0;
+    return NULL;
+  }
+  /* Helper to look for TEBs inside single allocated region since W10 1511. */
+  char *fill_if_match (char *start, char *dest)
+  {
+    for (region *r = regions; r; r = r->next)
+      if (r->teb && start == r->start)
+	{
+	  __small_sprintf (dest, "[teb (tid %ld)]", r->thread_id);
+	  return r->end;
+	}
+    return NULL;
   }
 
   ~thread_info ()
@@ -825,6 +837,7 @@ format_process_maps (void *data, char *&destbuf)
     char *abase;
     char *rbase;
     char *rend;
+    DWORD state;
   } cur = {{{'\0'}}, (char *)1, 0, 0};
 
   MEMORY_BASIC_INFORMATION mb;
@@ -838,6 +851,7 @@ format_process_maps (void *data, char *&destbuf)
   PMEMORY_SECTION_NAME msi = (PMEMORY_SECTION_NAME) tp.w_get ();
   char *posix_modname = tp.c_get ();
   size_t maxsize = 0;
+  char *peb_teb_abase = NULL;
 
   if (destbuf)
     {
@@ -847,12 +861,7 @@ format_process_maps (void *data, char *&destbuf)
 
   /* Iterate over each VM region in the address space, coalescing
      memory regions with the same permissions. Once we run out, do one
-     last_pass to trigger output of the last accumulated region.
-     
-     FIXME:  32 bit processes can't get address information beyond the
-	     32 bit address space from 64 bit processes.  We have to run
-	     this functionality in the target process, if the target
-	     process is 64 bit and our own process is 32 bit. */
+     last_pass to trigger output of the last accumulated region. */
   for (char *i = 0;
        VirtualQueryEx (proc, i, &mb, sizeof(mb)) || (1 == ++last_pass);
        i = cur.rend)
@@ -888,7 +897,8 @@ format_process_maps (void *data, char *&destbuf)
       region next = { a,
 		      (char *) mb.AllocationBase,
 		      (char *) mb.BaseAddress,
-		      (char *) mb.BaseAddress+mb.RegionSize
+		      (char *) mb.BaseAddress+mb.RegionSize,
+		      mb.State
       };
 
       /* Windows permissions are more fine-grained than the unix rwxp,
@@ -899,6 +909,57 @@ format_process_maps (void *data, char *&destbuf)
 	  cur.rend = next.rend; /* merge with previous */
       else
 	{
+	  char *peb_teb_end = NULL;
+peb_teb_rinse_repeat:
+	  /* Starting with W10 1511, PEB and TEBs don't get allocated
+	     separately.  Rather they are created in a single region.  Examine
+	     the region starting at the PEB address page-wise. */
+	  if (wincap.has_new_pebteb_region ())
+	    {
+	      if (peb_teb_abase && !peb_teb_end && cur.abase == peb_teb_abase)
+		{
+		  posix_modname[0] = '\0';
+		  peb_teb_end = cur.rend;
+		  if (cur.state == MEM_COMMIT)
+		    cur.rend = cur.rbase + wincap.page_size ();
+		}
+	      if (cur.state == MEM_COMMIT)
+		{
+		  if (!peb_teb_abase && cur.rbase == (char *) peb)
+		    {
+		      peb_teb_abase = cur.abase;
+		      peb_teb_end = cur.rend;
+		      cur.rend = cur.rbase + wincap.page_size ();
+		      strcpy (posix_modname, "[peb]");
+		    }
+		  else if (peb_teb_end)
+		    {
+		      char *end;
+		      posix_modname[0] = '\0';
+		      end = threads.fill_if_match (cur.rbase, posix_modname);
+
+		      if (end)
+			cur.rend = end;
+		      else
+			{
+			  char *base = cur.rbase;
+			  do
+			    {
+			      base += wincap.page_size ();
+			    }
+			  while (!threads.fill_if_match (base, posix_modname)
+				 && base < peb_teb_end);
+			  if (posix_modname[0])
+			    {
+			      posix_modname[0] = '\0';
+			      cur.rend = base;
+			    }
+			  else
+			      cur.rend = peb_teb_end;
+			}
+		    }
+		}
+	    }
 	  /* output the current region if it's "interesting". */
 	  if (cur.a.word)
 	    {
@@ -919,10 +980,18 @@ format_process_maps (void *data, char *&destbuf)
 	      len += written;
 	      len += __small_sprintf (destbuf + len, "%s\n", posix_modname);
 	    }
+
+	  if (peb_teb_end && cur.state == MEM_COMMIT)
+	    {
+	      cur.rbase = cur.rend;
+	      cur.rend += wincap.page_size ();
+	      if (cur.rbase < peb_teb_end)
+		goto peb_teb_rinse_repeat;
+	    }
 	  /* start of a new region (but possibly still the same allocation). */
 	  cur = next;
 	  /* if a new allocation, figure out what kind it is. */
-	  if (newbase && !last_pass && mb.State != MEM_FREE)
+	  if (newbase && !last_pass && cur.state != MEM_FREE)
 	    {
 	      /* If the return length pointer is missing, NtQueryVirtualMemory
 		 returns with STATUS_ACCESS_VIOLATION on Windows 2000. */
